import logging
from dataclasses import dataclass
from typing import Optional, List

import numpy as np
import pandas as pd
from lightgbm import LGBMRegressor

from player_performance_ratings.transformation.base_transformer import BaseTransformer


@dataclass
class ColumnWeight:
    name: str
    weight: float
    lower_is_better: bool = False


class ColumnsWeighter(BaseTransformer):

    def __init__(self,
                 weighted_column_name: str,
                 column_weights: list[ColumnWeight]
                 ):
        self.weighted_column_name = weighted_column_name
        self.column_weights = column_weights

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        df[self.weighted_column_name] = 0

        df['sum_cols_weights'] = 0
        for column_weight in self.column_weights:
            df[f'weight__{column_weight.name}'] = column_weight.weight
            df.loc[df[column_weight.name].isna(), f'weight__{column_weight.name}'] = 0
            df.loc[df[column_weight.name].isna(), column_weight.name] = 0
            df['sum_cols_weights'] = df['sum_cols_weights'] + df[f'weight__{column_weight.name}']

        drop_cols = ['sum_cols_weights']
        for column_weight in self.column_weights:
            df[f'weight__{column_weight.name}'] / df['sum_cols_weights']
            drop_cols.append(f'weight__{column_weight.name}')

        for column_weight in self.column_weights:

            if column_weight.lower_is_better:
                df[self.weighted_column_name] += df[f'weight__{column_weight.name}'] * (1 - df[column_weight.name])
            else:
                df[self.weighted_column_name] += df[f'weight__{column_weight.name}'] * df[column_weight.name]
        df = df.drop(columns=drop_cols)
        return df

    @property
    def features_created(self) -> list[str]:
        return [self.weighted_column_name]


class SklearnEstimatorImputer(BaseTransformer):

    def __init__(self, features: list[str], target_name: str, estimator: Optional[LGBMRegressor] = None, ):
        self.estimator = estimator or LGBMRegressor()
        self.features = features
        self.target_name = target_name

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self.estimator.fit(df[self.features], df[self.target_name])
        df = df.assign(**{
            f'imputed_col_{self.target_name}': self.estimator.predict(df[self.features])
        })
        df[self.target_name] = df[self.target_name].fillna(df[f'imputed_col_{self.target_name}'])
        return df.drop(columns=[f'imputed_col_{self.target_name}'])

    @property
    def features_created(self) -> list[str]:
        return [self.target_name]


class SkLearnTransformerWrapper(BaseTransformer):

    def __init__(self, transformer, features: list[str]):
        self.transformer = transformer
        self.features = features

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        df[self.features] = self.transformer.fit_transform(df[self.features])
        return df

    @property
    def features_created(self) -> list[str]:
        return self.features


class MinMaxTransformer(BaseTransformer):

    def __init__(self,
                 features: list[str],
                 quantile: float = 0.99,
                 allowed_mean_diff: Optional[float] = 0.01,
                 prefix: str = ""
                 ):
        self.features = features
        self.quantile = quantile
        self.allowed_mean_diff = allowed_mean_diff
        self.prefix = prefix
        if self.quantile < 0 or self.quantile > 1:
            raise ValueError("quantile must be between 0 and 1")

        self._feature_names_created = []

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        for feature in self.features:
            max_value = df[feature].quantile(self.quantile)
            min_value = df[feature].quantile(1 - self.quantile)

            df[self.prefix + feature] = (df[feature] - min_value) / (max_value - min_value)
            df[self.prefix + feature].clip(0, 1, inplace=True)
            self._feature_names_created.append(self.prefix + feature)

            if self.allowed_mean_diff:
                reps = 0
                mean_value = df[self.prefix + feature].mean()
                while abs(0.5 - mean_value) > self.allowed_mean_diff:

                    if mean_value > 0.5:
                        df[self.prefix + feature] = df[self.prefix + feature] * (1 - self.allowed_mean_diff)
                    else:
                        df[self.prefix + feature] = df[self.prefix + feature] * (1 + self.allowed_mean_diff)

                    df[self.prefix + feature].clip(0, 1, inplace=True)
                    mean_value = df[self.prefix + feature].mean()

                    reps += 1
                    if reps > 100:
                        logging.warning(
                            f"MinMaxTransformer: {feature} mean value is {mean_value} after {reps} repetitions")
                        continue

        return df

    @property
    def features_created(self) -> list[str]:
        return self._feature_names_created


class DiminishingValueTransformer(BaseTransformer):

    def __init__(self,
                 features: List[str],
                 cutoff_value: Optional[float] = None,
                 quantile_cutoff: float = 0.93,
                 excessive_multiplier: float = 0.8,
                 reverse: bool = False,
                 ):
        self.features = features
        self.cutoff_value = cutoff_value
        self.excessive_multiplier = excessive_multiplier
        self.trained_count: int = 0
        self.quantile_cutoff = quantile_cutoff
        self.reverse = reverse

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:

        for feature_name in self.features:
            if self.cutoff_value is None:
                cutoff_value = df[feature_name].quantile(self.quantile_cutoff)
            else:
                cutoff_value = self.cutoff_value

            if self.reverse:
                cutoff_value = 1 - cutoff_value
                df = df.assign(**{
                    feature_name: lambda x: np.where(
                        x[feature_name] <= cutoff_value,
                        - (cutoff_value - df[feature_name]) * self.excessive_multiplier + cutoff_value,
                        x[feature_name]

                    )
                })
            else:

                df = df.assign(**{
                    feature_name: lambda x: np.where(
                        x[feature_name] >= cutoff_value,
                        (x[feature_name] - cutoff_value).clip(lower=0) * self.excessive_multiplier + cutoff_value,
                        x[feature_name]
                    )
                })

        df[self.features] = df[self.features].fillna(df[self.features])

        return df

    @property
    def features_created(self) -> list[str]:
        return self.features


class SymmetricDistributionTransformer(BaseTransformer):

    def __init__(self, features: List[str], granularity: Optional[list[str]] = None, skewness_allowed: float = 0.1,
                 max_iterations: int = 15):
        self.features = features
        self.granularity = granularity
        self.skewness_allowed = skewness_allowed
        self.max_iterations = max_iterations

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:

        df = df.copy()

        for feature in self.features:
            if self.granularity:
                for column_granularity in self.granularity:
                    unique_values = df[column_granularity].unique()
                    for unique_value in unique_values:
                        rows = df[df[column_granularity] == unique_value]

                        rows = self._transform_rows(rows=rows, feature=feature)

                        condition = df[column_granularity] == unique_value
                        if isinstance(rows[feature], (list, np.ndarray, pd.Series)):
                            assert len(rows[
                                           feature]) == condition.sum(), "Length of rows[feature] must match the number of rows in condition"

                        df.loc[df[column_granularity] == unique_value, feature] = rows[feature]

            else:
                df = self._transform_rows(rows=df, feature=feature)
        return df

    def _transform_rows(self, rows: pd.DataFrame, feature: str) -> pd.DataFrame:
        skewness = rows[feature].skew()

        excessive_multiplier = 0.8
        quantile_cutoff = 0.95

        iteration = 0
        while abs(skewness) > self.skewness_allowed and len(
                rows) > 10 and iteration < self.max_iterations:

            if skewness < 0:
                reverse = True
            else:
                reverse = False
            diminishing_value_transformer = DiminishingValueTransformer(features=[feature],
                                                                        reverse=reverse,
                                                                        excessive_multiplier=excessive_multiplier,
                                                                        quantile_cutoff=quantile_cutoff)
            rows = diminishing_value_transformer.transform(rows)
            excessive_multiplier *= 0.96
            quantile_cutoff *= 0.993
            iteration += 1
            skewness = rows[feature].skew()

        return rows

    @property
    def features_created(self) -> list[str]:
        return self.features


class GroupByTransformer(BaseTransformer):

    def __init__(self,
                 features: list[str],
                 granularity: list[str],
                 agg_func: str = "mean",
                 prefix: str = "mean_",
                 ):
        self.features = features
        self.granularity = granularity
        self.agg_func = agg_func
        self.prefix = prefix
        self._feature_names_created = []
        for feature in self.features:
            self._feature_names_created.append(self.prefix + feature)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        for idx, feature in enumerate(self.features):
            df = df.assign(
                **{self._feature_names_created[idx]: df.groupby(self.granularity)[feature].transform(self.agg_func)})
        return df

    @property
    def features_created(self) -> list[str]:
        return self._feature_names_created


class NetOverPredictedTransformer(BaseTransformer):

    def __init__(self,
                 features: list[str],
                 granularity: list[str],
                 predict_transformer: Optional[BaseTransformer] = None,
                 prefix: str = "mean_",
                 ):
        self.granularity = granularity
        self.predict_transformer = predict_transformer or GroupByTransformer(features=features, granularity=granularity,
                                                                             agg_func='mean')
        self.prefix = prefix
        self.features = features
        self._feature_names_created = []
        for idx, predicted_feature in enumerate(self.predict_transformer.features_created):
            new_feature_name = f'{self.prefix}{self.features[idx]}'
            self._feature_names_created.append(new_feature_name)

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        df = df.assign(__id=range(1, len(df) + 1))
        predicted_df = self.predict_transformer.transform(df)
        df = df.merge(predicted_df[self.predict_transformer.features_created + ['__id']], on="__id").drop(
            columns=["__id"])

        for idx, predicted_feature in enumerate(self.predict_transformer.features_created):
            new_feature_name = self._feature_names_created[idx]
            df = df.assign(**{new_feature_name: df[self.features[idx]] - df[predicted_feature]})

        return df

    @property
    def features_created(self) -> list[str]:
        return self._feature_names_created



from typing import Union

import pandas as pd
from sklearn.preprocessing import StandardScaler

from player_performance_ratings import ColumnNames
from player_performance_ratings.transformation import ColumnWeight, DiminishingValueTransformer, \
    SkLearnTransformerWrapper, MinMaxTransformer, ColumnsWeighter

from player_performance_ratings.transformation.base_transformer import BaseTransformer
from player_performance_ratings.transformation.pre_transformers import GroupByTransformer, NetOverPredictedTransformer, \
    SymmetricDistributionTransformer


def auto_create_pre_transformers(column_weights: list[list[ColumnWeight]],
                                 column_names: list[ColumnNames]) -> list[
    BaseTransformer]:
    """
    Creates a list of transformers that ensure the performance column is generated in a way that makes sense for the rating model.
    Ensures columns aren't too skewed, scales them to similar ranges, ensure values are between 0 and 1,
    and then weights them according to the column_weights.
    """
    steps = []
    if not isinstance(column_weights[0], list):
        column_weights = [column_weights]

    all_feature_names = []
    for col_weights in column_weights:
        for col_weight in col_weights:
            if col_weight.name not in all_feature_names:
                all_feature_names.append(col_weight.name)

    feature_names = []

    for idx, col_weights in enumerate(column_weights):
        if column_names[idx].position is None:
            granularity = []
        else:
            granularity = [column_names[idx].position]

        if idx == 0 or column_names[idx].position != column_names[idx - 1].position:
            distribution_transformer = SymmetricDistributionTransformer(features=feature_names, granularity=granularity)
            steps.append(distribution_transformer)
        if column_names[idx].position is not None:
            feats = [c.name for c in column_weights[idx]]
            position_predicted_transformer = NetOverPredictedTransformer(features=feats,
                                                                         granularity=[column_names[idx].position])
            steps.append(position_predicted_transformer)

        for column_weight in col_weights:

            feature = column_weight.name

            if feature in feature_names:
                continue

            feature_names.append(feature)

    steps.append(
        SkLearnTransformerWrapper(transformer=StandardScaler(), features=all_feature_names))

    steps.append(MinMaxTransformer(features=all_feature_names))

    for idx, col_weights in enumerate(column_weights):
        column_weighter = ColumnsWeighter(weighted_column_name=column_names[idx].performance,
                                          column_weights=col_weights)
        steps.append(column_weighter)

    return steps



import logging
from typing import List, Optional, Union

import pandas as pd
import pendulum
from player_performance_ratings.transformation import ColumnWeight

from player_performance_ratings.consts import PredictColumnNames
from player_performance_ratings.predictor.estimators.base_estimator import BaseMLWrapper

from player_performance_ratings.predictor.estimators import SKLearnWrapper
from player_performance_ratings.data_structures import ColumnNames, Match
from player_performance_ratings.ratings.league_identifier import LeagueIdentifier
from player_performance_ratings.ratings.match_generator import convert_df_to_matches
from player_performance_ratings.ratings.opponent_adjusted_rating.rating_generator import RatingGenerator, \
    OpponentAdjustedRatingGenerator
from player_performance_ratings.transformation.base_transformer import BaseTransformer
from player_performance_ratings.transformation.factory import auto_create_pre_transformers


class MatchPredictor():

    def __init__(self,
                 column_names: Union[ColumnNames, list[ColumnNames]],
                 rating_generators: Optional[Union[RatingGenerator, list[RatingGenerator]]] = None,
                 pre_rating_transformers: Optional[List[BaseTransformer]] = None,
                 post_rating_transformers: Optional[List[BaseTransformer]] = None,
                 predictor: [Optional[BaseMLWrapper]] = None,
                 train_split_date: Optional[pendulum.datetime] = None,
                 use_auto_pre_transformers: bool = False,
                 column_weights: Optional[Union[list[list[ColumnWeight]], list[ColumnWeight]]] = None
                 ):

        """

        :param column_names:
        :param rating_generators: A single or a list of RatingGenerators.
        :param pre_rating_transformers: An optional list of transformations that take place rating generation.
            This is generally recommended if a more complex performance-value is used to update ratings.
            Although any type of feature engineering that isn't dependant upon the output of the ratings can be performed here.
        :param post_rating_transformers:
            After rating-generation, additional feature engineering can be performed.
        :param predictor:
        :param train_split_date:
        :param use_auto_pre_transformers:
            If true, the pre_rating_transformers will be automatically generated to ensure the performance-value is done according to good practices.
            For new users, this is recommended.
        :param column_weights:
            If auto_create_pre_transformers is True, column_weights must be set.
            It is generally used when multiple columns are used to calculate ratings and the columns need to be weighted when converting it to a performance_value.
            Even if only 1  feature is used but auto_create_pre_transformers is used,
             then it must still be created in order for auto_create_pre_transformers to know which columns needs to be transformed.


        """


        if rating_generators is None:
            rating_generators = [OpponentAdjustedRatingGenerator()]
            logging.warning(
                "rating generator not set. Uses OpponentAdjustedRatingGenerator. To run match_predictor without rating_generator set rating_generator to []")

        self.rating_generators = rating_generators if isinstance(rating_generators, list) else [rating_generators]
        self.column_names = column_names if isinstance(column_names, list) else [column_names for _ in
                                                                                 self.rating_generators]
        if len(rating_generators) == 0:
            if isinstance(column_names, list):
                self.column_names = column_names
            else:
                self.column_names = [column_names]

        self.use_auto_pre_transformers = use_auto_pre_transformers
        if self.use_auto_pre_transformers and not column_weights:
            raise ValueError("column_weights must be set if auto_create_pre_transformers is True")

        self.column_weights = column_weights
        if self.column_weights and isinstance(self.column_weights[0], ColumnWeight):
            self.column_weights = [self.column_weights]

        if not self.use_auto_pre_transformers and column_weights:
            logging.warning(
                "column_weights is set but auto_create_pre_transformers is False. column_weights will be ignored")



        self.pre_rating_transformers = pre_rating_transformers or []
        if self.use_auto_pre_transformers:
            self.pre_rating_transformers = auto_create_pre_transformers(column_weights=self.column_weights,
                                                                     column_names=self.column_names)

        self.post_rating_transformers = post_rating_transformers or []

        self.predictor = predictor

        if self.predictor is None:
            features = []

            for c in self.post_rating_transformers:
                features += c.features_created
            for c in self.rating_generators:
                features += c.features_created

            logging.warning(f"predictor is not set. Will use {features} as features")

            self.predictor = SKLearnWrapper(
                features=features,
                target=PredictColumnNames.TARGET
            )

        self.predictor.set_target(PredictColumnNames.TARGET)
        self.train_split_date = train_split_date

    def generate_historical(self, df: pd.DataFrame, matches: list[Match] = None,
                            store_ratings: bool = True) -> pd.DataFrame:

        if self.predictor.target not in df.columns:
            raise ValueError(
                f"Target {self.predictor.target} not in df columns. Target always needs to be set equal to {PredictColumnNames.TARGET}")


        for pre_rating_transformer in self.pre_rating_transformers:
            df = pre_rating_transformer.transform(df)

        if self.train_split_date is None:
            self.train_split_date = df.iloc[int(len(df) / 1.3)][self.column_names[0].start_date]

        for rating_idx, rating_generator in enumerate(self.rating_generators):

            rating_column_names = self.column_names[rating_idx]

            if matches is None:
                matches = convert_df_to_matches(column_names=rating_column_names, df=df,
                                                league_identifier=LeagueIdentifier())

            if store_ratings:
                match_ratings = rating_generator.generate(matches, df=df, column_names=rating_column_names)
            else:
                match_ratings = rating_generator.generate(matches)
            for rating_feature, values in match_ratings.items():
                if len(self.rating_generators) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature
                df[rating_feature_str] = values

        for post_rating_transformer in self.post_rating_transformers:
            df = post_rating_transformer.transform(df)

        train_df = df[df[self.column_names[0].start_date] <= self.train_split_date]

        self.predictor.train(train_df)
        df = self.predictor.add_prediction(df)
        return df

    def predict(self, df: pd.DataFrame) -> pd.DataFrame:
        for pre_rating_transformer in self.pre_rating_transformers:
            df = pre_rating_transformer.transform(df)

        for rating_idx, rating_generator in enumerate(self.rating_generators):
            rating_column_names = self.column_names[rating_idx]

            matches = convert_df_to_matches(column_names=rating_column_names, df=df,
                                            league_identifier=LeagueIdentifier())

            match_ratings = rating_generator.generate(matches, df=df)
            for rating_feature, values in match_ratings.items():

                if len(self.rating_generators) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature
                df[rating_feature_str] = values

        for post_rating_transformer in self.post_rating_transformers:
            df = post_rating_transformer.transform(df)

        df = self.predictor.add_prediction(df)
        return df

    @property
    def classes_(self):
        return self.predictor.model.classes_
